import click
import subprocess
import configparser
import json
import jinja2
import os
import time
import asyncio
import typing
import uuid
import sqlalchemy.exc
from tabulate import tabulate
from typing import Optional, List
import csv
from joule import utilities
from aiohttp.test_utils import unused_port
from joule import errors, api

TEMPLATE_DIR = os.path.join(os.path.dirname(__file__), 'local_postgres_templates')

if typing.TYPE_CHECKING:
    import sqlalchemy
    from sqlalchemy.orm import Session
    from joule.models import Base, Stream, folder, DataStore, TimescaleStore, NilmdbStore


@click.command(name="ingest")
@click.option("-c", "--config", help="main configuration file", default="/etc/joule/main.conf")
@click.option("-b", "--backup", help="backup to ingest")
@click.option("-n", "--node", help="node to ingest")
@click.option("-m", "--map", help="map file of source to destination streams")
@click.option("-p", "--pgctl-binary", help="override default pg_ctl location")
@click.option("-y", "--yes", help="do not ask for confirmation", is_flag=True)
@click.option('-s', "--start", help="timestamp or descriptive string")
@click.option('-e', "--end", help="timestamp or descriptive string")
def admin_ingest(config, backup, node, map, pgctl_binary, yes, start, end):
    # expensive imports so only execute if the function is called
    from joule.services import load_config
    import sqlalchemy
    from sqlalchemy.orm import Session
    from joule.models import Base, TimescaleStore, NilmdbStore

    parser = configparser.ConfigParser()
    loop = asyncio.get_event_loop()
    # make sure either a backup or a node is specified
    if (((backup is None) and (node is None)) or
            ((backup is not None) and (node is not None))):
        raise click.ClickException("Specify either a backup or a node to ingest data from")

    # make sure the time bounds make sense
    if start is not None:
        try:
            start = utilities.human_to_timestamp(start)
        except ValueError:
            raise errors.ApiError("invalid start time: [%s]" % start)
    if end is not None:
        try:
            end = utilities.human_to_timestamp(end)
        except ValueError:
            raise errors.ApiError("invalid end time: [%s]" % end)
    if (start is not None) and (end is not None) and ((end - start) <= 0):
        raise click.ClickException("Error: start [%s] must be before end [%s]" % (
            utilities.timestamp_to_human(start),
            utilities.timestamp_to_human(end)))

    # parse the map file if specified
    stream_map = None
    if map is not None:
        stream_map = []
        try:
            with open(map, newline='') as csvfile:
                reader = csv.reader(csvfile, delimiter=',', quotechar='|', skipinitialspace=True)
                for row in reader:
                    if len(row) == 0:  # ignore blank lines
                        continue
                    if len(row) == 1 and len(row[0]) == 0:  # line with only whitespace
                        continue
                    if row[0][0] == '#':  # ignore comments
                        continue
                    if len(row) != 2:
                        raise errors.ConfigurationError("""invalid map format. Refer to template below:
    
     # this line is a comment
     # only paths in this file will be copied
     # source and destination paths are separated by a ','
     
     /source/path, /destination/path
     /source/path2, /destination/path2
     #..etc
     
     """)
                    stream_map.append(row)
        except FileNotFoundError:
            raise click.ClickException("Cannot find map file at [%s]" % map)
        except PermissionError:
            raise click.ClickException("Cannot read map file at [%s]" % map)
        except errors.ConfigurationError as e:
            raise click.ClickException(str(e))

    # load the Joule configuration file
    try:
        with open(config, 'r') as f:
            parser.read_file(f, config)
            joule_config = load_config.run(custom_values=parser)
    except FileNotFoundError:
        raise click.ClickException("Cannot load joule configuration file at [%s]" % config)
    except PermissionError:
        raise click.ClickException("Cannot read joule configuration file at [%s] (run as root)" % config)
    except errors.ConfigurationError as e:
        raise click.ClickException("Invalid configuration: %s" % e)

    dest_engine = sqlalchemy.create_engine(joule_config.database)

    Base.metadata.create_all(dest_engine)
    dest_db = Session(bind=dest_engine)
    if joule_config.nilmdb_url is not None:
        dest_datastore = NilmdbStore(joule_config.nilmdb_url, 0, 0, loop)
    else:
        dest_datastore = TimescaleStore(joule_config.database, 0, 0, loop)

    # demote priveleges
    if "SUDO_GID" in os.environ:
        os.setgid(int(os.environ["SUDO_GID"]))
    if "SUDO_UID" in os.environ:
        os.setuid(int(os.environ["SUDO_UID"]))

    # create a log file for exec cmds
    pg_log_name = "joule_restore_log_%s.txt" % uuid.uuid4().hex.upper()[0:6]
    pg_log = open(pg_log_name, 'w')

    # if pgctl_binary is not specified, try to autodect it
    if pgctl_binary is None:
        try:
            completed_proc = subprocess.run(["psql", "-V"], stdout=subprocess.PIPE)
            output = completed_proc.stdout.decode('utf-8')
            version = output.split(" ")[2]
            major_version = version.split(".")[0]
            pgctl_binary = "/usr/lib/postgresql/%s/bin/pg_ctl" % major_version
        except (FileNotFoundError, IndexError):
            raise click.ClickException("cannot autodetect pg_ctl location, specify with -b")

    # determine if the source is a backup or a node
    if node is not None:
        live_restore = True
        src_dsn = loop.run_until_complete(get_dsn(node))
        # check whether the source uses nilmdb
        click.echo("WARNING: Nilmdb sources are not supported yet")
        src_datastore = TimescaleStore(src_dsn, 0, 0, loop)
        nilmdb_proc = None
    else:
        if not os.path.isdir(backup):
            raise click.ClickException("backup [%s] does not exist" % backup)
        src_dsn = start_src_db(backup, pgctl_binary, pg_log)
        # check whether the source uses nilmdb
        nilmdb_path = os.path.join(backup, 'nilmdb')
        nilmdb_proc = None
        if os.path.exists(nilmdb_path):
            port = unused_port()
            nilmdb_proc = start_src_nilmdb(nilmdb_path, port, pg_log)
            click.echo("waiting for nilmdb to initialize...")
            time.sleep(2)
            src_datastore = NilmdbStore('http://127.0.0.1:%d' % port, 0, 0, loop)
            if joule_config.nilmdb_url is None:
                click.echo("Note: re-copying from NilmDB to Timescale may result in --nothing to copy-- messages")
        else:
            src_datastore = TimescaleStore(src_dsn, 0, 0, loop)

        live_restore = False

    src_engine = sqlalchemy.create_engine(src_dsn)

    num_tries = 0
    max_tries = 1
    while True:
        try:
            Base.metadata.create_all(src_engine)
            break
        except sqlalchemy.exc.OperationalError as e:
            if live_restore:
                raise click.ClickException(str(e))  # this should work immediately
            num_tries += 1
            click.echo("... attempting to connect to source database (%d/%d)" % (num_tries, max_tries))
            time.sleep(2)
            if num_tries >= max_tries:
                raise click.ClickException("cannot connect to source database, log saved in [%s]" % pg_log_name)

    src_db = Session(bind=src_engine)

    try:
        loop.run_until_complete(run(src_db, dest_db,
                                    src_datastore, dest_datastore,
                                    stream_map, yes,
                                    start, end))
    except errors.ConfigurationError as e:
        print("Logs written to [%s]" % pg_log_name)
        raise click.ClickException(str(e))
    finally:
        # close connections
        dest_db.close()
        src_db.close()
        loop.run_until_complete(dest_datastore.close())
        loop.run_until_complete(src_datastore.close())
        # clean up database if not a live_restore
        if not live_restore:
            args = ["-D", os.path.join(backup)]
            args += ["stop"]
            cmd = [pgctl_binary] + args
            subprocess.call(cmd, stderr=pg_log, stdout=pg_log)
            sock_path = os.path.join(backup, 'sock')
            sockets = os.listdir(sock_path)
            for s in sockets:
                os.remove(os.path.join(sock_path, s))
            os.rmdir(sock_path)
            if nilmdb_proc is not None:
                nilmdb_proc.terminate()
                nilmdb_proc.communicate()

    pg_log.close()
    os.remove(pg_log_name)
    click.echo("OK")


async def get_dsn(node_name) -> str:
    node = api.get_node(node_name)
    conn_info = await node.db_connection_info()
    await node.close()
    return conn_info.to_dsn()


def start_src_db(backup_path, pgctl_binary, log) -> str:
    # make sure file permissions are correct
    os.chmod(backup_path, 0o700)
    # read the info file for database name and user
    with open(os.path.join(backup_path, "info.json"), 'r') as f:
        db_info = json.load(f)

    # create the config files
    env = jinja2.Environment(loader=jinja2.FileSystemLoader(TEMPLATE_DIR))

    template = env.get_template("postgresql.conf.jinja2")
    sock_path = os.path.join(backup_path, "sock")

    # remove sockets if they exist (from a previous use of this backup)
    if os.path.isdir(sock_path):
        click.confirm("It looks like another ingest either failed or is still running, continue anyway?", abort=True)
        sockets = os.listdir(sock_path)
        for s in sockets:
            os.remove(os.path.join(sock_path, s))
        os.rmdir(sock_path)
        if os.path.exists(os.path.join(backup_path, 'postmaster.pid')):
            os.remove(os.path.join(backup_path, 'postmaster.pid'))
    os.mkdir(sock_path)
    db_port = unused_port()
    output = template.render(port=db_port, sock_dir=os.path.abspath(sock_path))
    with open(os.path.join(backup_path, "postgresql.conf"), "w") as f:
        f.write(output)

    template = env.get_template("pg_hba.conf.jinja2")
    output = template.render(user=db_info["user"])
    with open(os.path.join(backup_path, "pg_hba.conf"), "w") as f:
        f.write(output)

    template = env.get_template("pg_ident.conf.jinja2")
    output = template.render()
    with open(os.path.join(backup_path, "pg_ident.conf"), "w") as f:
        f.write(output)

    template = env.get_template("recovery.conf.jinja2")
    wal_dir = os.path.join(os.path.abspath(backup_path), 'pg_wal')
    output = template.render(wal_path=wal_dir)
    with open(os.path.join(backup_path, "recovery.conf"), "w") as f:
        f.write(output)

    # start postgres

    args = ["-D", backup_path]
    args += ["start"]
    cmd = [pgctl_binary] + args
    try:
        subprocess.call(cmd, stderr=log, stdout=log)
    except FileNotFoundError:
        raise click.ClickException(
            "Cannot find pg_ctl, expected [%s] to exist. Specify location with -b" % pgctl_binary)

    click.echo("waiting for postgres to initialize")
    time.sleep(2)

    # connect to the database
    return "postgresql://%s:%s@localhost:%d/%s" % (
        db_info["user"],
        db_info["password"],
        db_port,
        db_info["database"])


def start_src_nilmdb(database_path, port, log) -> subprocess.Popen:
    args = ["--address", "127.0.0.1"]
    args += ["--port", "%s" % port]
    args += ["--database", database_path]
    cmd = ['nilmdb-server'] + args
    return subprocess.Popen(cmd, stderr=log, stdout=log)


async def run(src_db: 'Session',
              dest_db: 'Session',
              src_datastore: 'DataStore',
              dest_datastore: 'DataStore',
              stream_map: Optional[List],
              confirmed: bool,
              start: Optional[int],
              end: Optional[int]):
    from joule.models import Stream, folder, stream
    from joule.services import parse_pipe_config

    src_streams = src_db.query(Stream).all()
    dest_streams = dest_db.query(Stream).all()
    await src_datastore.initialize(src_streams)
    await dest_datastore.initialize(dest_streams)

    if stream_map is None:
        src_streams = src_db.query(Stream).all()
        src_paths = map(folder.get_stream_path, src_streams)
        stream_map = map(lambda _path: [_path, _path], src_paths)

    # create the copy map array
    copy_maps = []
    for item in stream_map:
        # get the source stream
        source = folder.find_stream_by_path(item[0], src_db)
        if source is None:
            raise errors.ConfigurationError("source stream [%s] does not exist" % item[0])
        src_intervals = await src_datastore.intervals(source, start, end)
        # get or create the destination stream
        dest = folder.find_stream_by_path(item[1], dest_db)
        if dest is None:
            (path, name, _) = parse_pipe_config.parse_pipe_config(item[1])
            dest_folder = folder.find(path, dest_db, create=True)
            dest = stream.from_json(source.to_json())
            # set the attributes on the new stream
            dest.name = name
            dest.keep_us = dest.KEEP_ALL
            dest.is_configured = False
            dest.is_source = False
            dest.is_destination = False
            dest.id = None
            for e in dest.elements:
                e.id = None
            dest_folder.streams.append(dest)
            dest_intervals = None
        else:
            # make sure the destination is compatible
            if dest.layout != source.layout:
                raise errors.ConfigurationError(
                    "source stream [%s] is not compatible with destination stream [%s]" % (item[0], item[1]))

            dest_intervals = await dest_datastore.intervals(dest, start, end)
        # figure out the time bounds to copy
        if dest_intervals is None:
            copy_intervals = src_intervals
        else:
            copy_intervals = utilities.interval_difference(src_intervals, dest_intervals)

        copy_maps.append(CopyMap(source, dest, copy_intervals))

    # display the copy table
    rows = []
    copy_required = False
    for item in copy_maps:
        if item.start is None:
            start = "\u2014"
            end = "\u2014"
        else:
            start = utilities.timestamp_to_human(item.start)
            end = utilities.timestamp_to_human(item.end)
            copy_required = True
        rows.append([item.source_path, item.dest_path, start, end])
    click.echo(tabulate(rows,
                        headers=["Source", "Destination", "From", "To"],
                        tablefmt="fancy_grid"))
    if not copy_required:
        click.echo("No data needs to be copied")
        return

    if not confirmed and not click.confirm("Start data copy?"):
        click.echo("cancelled")
        return

    dest_db.commit()
    # execute the copy
    for item in copy_maps:
        await copy(item, src_datastore, dest_datastore, src_db, dest_db)


num_rows = 0


async def copy(copy_map: 'CopyMap',
               src_datastore: 'DataStore',
               dest_datastore: 'DataStore',
               src_db: 'Session',
               dest_db: 'Session'):
    from joule.models import annotation, Annotation
    global num_rows
    num_rows = 0
    # compute the duration of data to copy
    duration = 0
    for interval in copy_map.intervals:
        duration += interval[1] - interval[0]

    with click.progressbar(
            label='[%s] --> [%s]' % (copy_map.source_path, copy_map.dest_path),
            length=duration) as bar:
        for interval in copy_map.intervals:
            await copy_interval(interval[0], interval[1], bar,
                                copy_map.source, copy_map.dest,
                                src_datastore, dest_datastore)
            start_dt = utilities.timestamp_to_datetime(interval[0])
            end_dt = utilities.timestamp_to_datetime(interval[1])
            # remove existing annotations (if any)
            dest_db.query(Annotation). \
                filter(Annotation.stream == copy_map.dest). \
                filter(Annotation.start >= start_dt). \
                filter(Annotation.start < end_dt). \
                delete()
            # retrieve source annotations that start in this interval
            items: List[Annotation] = src_db.query(Annotation). \
                filter(Annotation.stream == copy_map.source). \
                filter(Annotation.start >= start_dt). \
                filter(Annotation.start < end_dt)
            # copy them over to the destination
            for item in items:
                item_copy = annotation.from_json(item.to_json())
                item_copy.id = None
                item_copy.stream = copy_map.dest
                dest_db.add(item_copy)
            dest_db.commit()
    if num_rows == 0:
        print("[%s]\t--nothing to copy--" % copy_map.dest_path)
    else:
        print("[%s]\t copied %d rows" % (copy_map.dest_path, num_rows))


async def copy_interval(start: int, end: int, bar,
                        src_stream: 'Stream', dest_stream: 'Stream',
                        src_datastore: 'DataStore', dest_datastore: 'DataStore'):
    from joule.models import pipes, Stream
    pipe = pipes.LocalPipe(src_stream.layout, write_limit=4, debug=False)
    dest_stream.keep_us = Stream.KEEP_ALL  # do not delete any data
    insert_task = await dest_datastore.spawn_inserter(dest_stream,
                                                      pipe, asyncio.get_event_loop())

    last_ts = start
    async def writer(data, layout, decimated):
        nonlocal last_ts
        global num_rows

        num_rows += (len(data))
        if len(data) > 0:
            cur_ts = data['timestamp'][-1]
            await pipe.write(data)
            # await asyncio.sleep(0.01)
            bar.update(cur_ts - last_ts)
            last_ts = cur_ts

    await src_datastore.extract(src_stream, start, end, writer)
    await pipe.close()
    await insert_task

    bar.update(end - last_ts)


class CopyMap:
    def __init__(self, source: 'Stream', dest: 'Stream', intervals: List):
        self.source = source
        self.dest = dest
        self.intervals = intervals
        if len(intervals) > 0:
            self.start = intervals[0][0]
            self.end = intervals[-1][1]
        else:
            self.start = None
            self.end = None

    @property
    def source_path(self) -> str:
        from joule.models import folder
        if self.source is None:
            return "--none--"
        return folder.get_stream_path(self.source)

    @property
    def dest_path(self) -> str:
        from joule.models import folder
        if self.dest is None:
            return "--none--"
        return folder.get_stream_path(self.dest)

    def __str__(self):
        return "[%s] --> [%s] [%d intervals]" % (
            self.source_path,
            self.dest_path,
            len(self.intervals))



import click
import asyncio

from joule import errors
from joule.cli.config import Config, pass_config
from joule.api import BaseNode


@click.command(name="move")
@click.argument("source")
@click.argument("destination")
@pass_config
def move(config: Config, source: str, destination: str):
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            config.node.folder_move(source, destination))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()
    click.echo("OK")


@click.command(name="rename")
@click.argument("folder")
@click.argument("name")
@pass_config
def rename(config: Config, folder, name):
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            _run_rename(config.node, folder, name))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()
    click.echo("OK")


async def _run_rename(node: BaseNode, folder_path: str, name: str):
    folder = await node.folder_get(folder_path)
    folder.name = name
    await node.folder_update(folder)


@click.command(name="delete")
@click.option("--recursive", "-r", is_flag=True)
@click.argument("folder")
@pass_config
def delete(config, folder, recursive):
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            config.node.folder_delete(folder, recursive))
        click.echo("OK")
    except errors.ApiError as e:
        raise click.ClickException(str(e))
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


@click.group(name="folder")
def folders():
    """Manage Joule folders"""
    pass  # pragma: no cover


folders.add_command(move)
folders.add_command(delete)
folders.add_command(rename)



import click
import asyncio
from tabulate import tabulate

from joule import errors
from joule.api import BaseNode
from joule.cli.config import pass_config


@click.command(name="list")
@pass_config
def cli_list(config):
    loop = asyncio.get_event_loop()

    try:
        loop.run_until_complete(
            _run(config.node))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


async def _run(node: BaseNode):
    proxies = await node.proxy_list()
    # display module information
    headers = ['ID', 'Name', 'Target URL']
    result = []
    for proxy in proxies:
        data = [proxy.id, proxy.name, proxy.target_url]
        result.append(data)
    click.echo(tabulate(result,
                        headers=headers,
                        tablefmt="fancy_grid"))



import click
import asyncio

from joule import errors
from joule.cli.config import pass_config
from joule.api import BaseNode


@click.command(name="add")
@click.argument("type", type=click.Choice(['user', 'joule', 'lumen']))
@click.argument("identifier")
@pass_config
def cli_add(config, type, identifier):
    loop = asyncio.get_event_loop()
    try:
        if type == 'user':
            coro = _add_user(config.node, identifier)
        elif type == 'joule':
            coro = _add_node(config.node, identifier)
        elif type == 'lumen':
            coro = _add_lumen(config.node, identifier)
        else:
            raise click.ClickException("invalid type [%s]" % type)
        loop.run_until_complete(coro)
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


async def _add_lumen(node: BaseNode, host):
    # add the node and see what error comes back
    req_credentials = "unset"
    try:
        await node.master_add("lumen", host, {})
    except errors.ApiError as e:
        if "auth_key" in str(e):
            req_credentials = "key"
        elif "first_name" in str(e):
            req_credentials = "user"
        else:
            raise e
    if req_credentials == "unset":
        raise errors.ApiError("Lumen node did not specify required credentials")
    lumen_params = {}
    if req_credentials == "key":
        key = click.prompt("Lumen authorization key")
        lumen_params = {"auth_key": key}
    else:
        click.echo("This a new Lumen instance, create a user account")
        first_name = click.prompt("First Name")
        last_name = click.prompt("Last Name")
        email = click.prompt("E-mail")
        password = click.prompt("Password", hide_input=True, confirmation_prompt=True)
        lumen_params = {
            "first_name": first_name,
            "last_name": last_name,
            "email": email,
            "password": password}
    await node.master_add("lumen", host, lumen_params)
    click.echo("Access to [%s] granted to Lumen Node [%s]" % (node.name, host))
    if req_credentials == "user":
        click.echo("Log into the Lumen server using your e-mail and password")


async def _add_node(node: BaseNode, host):
    result = await node.master_add("joule", host)
    click.echo("Access to [%s] granted to Joule Node [%s]" % (node.name, result.name))


async def _add_user(node: BaseNode, name):
    result = await node.master_add("user", name)
    click.echo("Access to node [%s] granted to user [%s]" % (node.name, name))
    click.echo("")
    click.echo("Key:\t%s" % result.key)
    click.echo("")
    click.echo("Run the following command to install the key on the user's machine")
    click.echo("\t$> joule node add %s %s %s" % (node.name, node.url, result.key))
    click.echo("* IP address may differ based on network setup")



from typing import Dict, List, Callable
import logging
import asyncio
import shlex
import os
import json
import collections
import datetime
import psutil
import numpy as np

from joule.models.module import Module
from joule.models.stream import Stream
from joule.models.folder import get_stream_path
from joule.errors import SubscriptionError
from joule.models import pipes
from joule.models.pipes.errors import EmptyPipe, PipeError

# custom types
Loop = asyncio.AbstractEventLoop
Subscribers = Dict[Stream, List[pipes.Pipe]]

popen_lock = None
log = logging.getLogger('joule')

SOCKET_BASE = "/tmp/joule/module%d"
API_SOCKET = "/tmp/joule/api"


def _initialize_popen_lock():
    global popen_lock
    if popen_lock is None:
        popen_lock = asyncio.Lock()


class DataConnection:
    def __init__(self, name: str, child_fd: int,
                 stream: Stream, pipe: pipes.Pipe,
                 unsubscribe: Callable = None):
        self.name = name
        self.child_fd = child_fd
        self.stream = stream
        self.pipe = pipe
        self.unsubscribe = unsubscribe

    @property
    def location(self):
        return get_stream_path(self.stream)

    async def disconnect(self):
        if self.unsubscribe is not None:
            self.unsubscribe()
        await self.pipe.close()


class Statistics:
    def __init__(self, pid, create_time, cpu_percent, memory_percent):
        self.pid = pid
        self.create_time = create_time
        self.cpu_percent = cpu_percent
        self.memory_percent = memory_percent

    def to_json(self):
        return {
            'pid': self.pid,
            'create_time': self.create_time,
            'cpu_percent': self.cpu_percent,
            'memory_percent': self.memory_percent
        }


class Worker:

    def __init__(self, my_module: Module):

        self.module: Module = my_module
        # map of subscribers (1-many) that consume module outputs
        self.subscribers: Subscribers = {}
        # map of (fd,pipe) connections to module input names
        self.input_connections: List[DataConnection] = []
        # map of (fd,pipe) connections to module output names
        self.output_connections: List[DataConnection] = []

        for (name, stream) in self.module.outputs.items():
            # add a subscriber array and an empty output connection
            self.subscribers[stream] = []

        self._logs = collections.deque([], maxlen=my_module.log_size)
        self.process: asyncio.subprocess.Process = None
        self.stop_requested = False

        # tunable constants
        # how long to wait for proc to stop nicely
        self.SIGTERM_TIMEOUT = 2
        # how long to wait to restart a failed process
        self.RESTART_INTERVAL = 3
        # how to wait for a subscriber to accept data
        self.SUBSCRIBER_TIMEOUT = 1
        # how long to try restarting if worker is currently missing inputs
        self.RESTART_AFTER_MISSING_INPUTS = 5

    async def statistics(self) -> Statistics:
        # gather process statistics
        try:
            if self.process is not None:
                p = psutil.Process(pid=self.process.pid)
                # collect cpu usage over a 0.5 second interval
                p.cpu_percent()
                await asyncio.sleep(0.5)
                with p.oneshot():
                    return Statistics(p.pid,
                                      p.create_time(),
                                      p.cpu_percent(),
                                      p.memory_percent())
            else:
                # worker is not running, no statistics available
                return Statistics(None, None, None, None)

        except psutil.NoSuchProcess:
            return Statistics(None, None, None, None)

    # provide the module attributes
    @property
    def uuid(self):
        return self.module.uuid

    @property
    def name(self):
        return self.module.name

    @property
    def description(self):
        return self.module.description

    @property
    def is_app(self):
        return self.module.is_app

    @property
    def interface_socket(self):
        if self.module.is_app:
            return SOCKET_BASE % self.module.uuid
        return None

    @property
    def interface_name(self):
        if self.module.is_app:
            return SOCKET_BASE % self.module.uuid
        return "none"

    def produces(self, stream: Stream) -> bool:
        # returns True if this worker produces [stream]
        for output in self.module.outputs.values():
            if output == stream:
                return True
        else:
            return False

    async def run(self, subscribe: Callable[[Stream, pipes.Pipe, Loop], Callable],
                  loop: Loop, restart: bool = True) -> None:
        self.stop_requested = False
        while True:
            # when jouled is run from the command line Ctrl+C sends SIGTERM
            # to the child and jouled which can cause jouled to restart the
            # child and then kill it
            if self.stop_requested:
                break  # pragma: no cover
            self.log("---starting module---")
            try:
                await self._spawn_child(subscribe, loop)
            except SubscriptionError as e:
                log.error("Cannot start module [%s]: %s" % (self.module.name, e))
                self.log("inputs are not available: %s" % e)
                break
            self.log("---module terminated---")
            if restart:
                await asyncio.sleep(self.RESTART_INTERVAL)
                if self.stop_requested:
                    break
                log.error("Restarting failed module: %s" % self.module.name)
                # insert an empty block in output_queues to indicate end of
                # interval
                for subscriber in self.subscribers.values():
                    for pipe in subscriber:
                        pipe.close_interval_nowait()

            else:
                break

    async def restart(self, loop: Loop) -> None:
        await self._stop_child(loop)

    async def stop(self, loop: Loop) -> None:
        self.stop_requested = True
        await self._stop_child(loop)

    async def _stop_child(self, loop: Loop) -> None:
        if self.process is None:
            return
        try:
            self.process.terminate()
        except ProcessLookupError:
            return  # process is already terminated
        try:
            await asyncio.wait_for(self.process.wait(),
                                   timeout=self.SIGTERM_TIMEOUT,
                                   loop=loop)
        except asyncio.TimeoutError:
            log.warning(
                "Cannot stop %s with SIGTERM, killing process" % self.module.name)
            try:
                self.process.kill()
            except ProcessLookupError:  # pragma: no cover
                pass  # if the process stopped after the timeout

    def log(self, msg):
        timestamp = datetime.datetime.now().isoformat()
        self._logs.append("[%s]: %s" % (timestamp, msg))
        # print("[%s: %s] " % (self.module.name, pid) + msg)

    @property
    def logs(self) -> List[str]:
        return list(self._logs)

    # returns a queue and unsubscribe function
    def subscribe(self, stream: Stream, pipe: pipes.Pipe) -> Callable:
        try:
            self.subscribers[stream].append(pipe)
        except KeyError:
            raise SubscriptionError()

        def unsubscribe():
            try:
                i = self.subscribers[stream].index(pipe)
                del self.subscribers[stream][i]
            except ValueError:
                # subscription already cancelled
                # this can happen if the _output_handler removes the subscriber
                pass

        return unsubscribe

    async def _spawn_child(self, subscribe: Callable[[Stream, pipes.Pipe, Loop], Callable],
                           loop: Loop) -> None:

        # lock so fd's don't pollute other modules
        _initialize_popen_lock()
        await popen_lock.acquire()
        try:

            await self._subscribe_to_inputs(subscribe, loop)
        except SubscriptionError as e:
            self._close_child_fds()
            popen_lock.release()
            await self._close_connections()
            raise e  # bubble up the exception

        output_task = await self._spawn_outputs(loop)
        cmd = self._compose_cmd()
        env = {**os.environ, 'PYTHONUNBUFFERED': '1'}
        create = asyncio.create_subprocess_exec(
            *cmd,
            stdin=asyncio.subprocess.DEVNULL,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
            env=env, close_fds=False)
        try:
            self.process = await create
        except Exception as e:
            self.process = None
            self.log("ERROR: cannot start module: \n\t%s" % e)
            self.module.status = Module.STATUS.FAILED
            log.error("Cannot start [%s]: %s" % (self.module.name, e))

            self._close_child_fds()
            popen_lock.release()
            await self._close_connections()
            output_task.cancel()
            try:
                await output_task
            # should be caught but on fast fails they can bubble up
            except asyncio.CancelledError:  # pragma: no cover
                pass
            return
        self._close_child_fds()
        popen_lock.release()

        logger_task = loop.create_task(self._logger())
        await self.process.wait()
        # Unwind the tasks
        await self._close_connections()
        output_task.cancel()
        logger_task.cancel()
        # collect any errors
        try:
            await output_task
            await logger_task
        # should be caught but on fast fails they can bubble up
        except asyncio.CancelledError:  # pragma: no cover
            pass

    async def _logger(self):
        try:
            stream = self.process.stdout
            while True:
                bline = await stream.readline()
                if len(bline) == 0:
                    break
                line = bline.decode('UTF-8').rstrip()
                self.log(line)
        except asyncio.CancelledError:  # pragma: no cover
            return

    async def _spawn_outputs(self, loop: Loop) -> asyncio.Task:
        tasks: List[asyncio.Task] = []
        # configure output pipes          [module]==>[worker]
        for (name, stream) in self.module.outputs.items():
            (r, w) = os.pipe()
            rf = pipes.reader_factory(r, loop)
            os.set_inheritable(w, True)
            pipe = pipes.InputPipe(name=name, stream=stream,
                                   reader_factory=rf)
            self.output_connections.append(DataConnection(
                name, w, stream, pipe))
            tasks.append(loop.create_task(
                self._output_handler(pipe, self.subscribers[stream], loop)))

        return asyncio.gather(*tasks)

    async def _output_handler(self, child_output: pipes.Pipe,
                              subscribers: List[pipes.Pipe], loop: Loop):
        """given a numpy pipe, get data and put it
           into each queue in [output_queues] """
        last_ts = None
        try:
            while True:
                data = await child_output.read()
                if len(data) > 0:

                    if not self._verify_monotonic_timestamps(data, last_ts, child_output.name):
                        for pipe in subscribers:
                            await pipe.close_interval()
                        await self.restart(loop)
                        break
                    last_ts = data['timestamp'][-1]

                    child_output.consume(len(data))
                    for pipe in subscribers[:]:
                        # if child_output.name=='output2':
                        #    print("writing to %d subscribers" % len(subscribers))
                        try:

                            await asyncio.wait_for(pipe.write(data),
                                                   self.SUBSCRIBER_TIMEOUT)
                        except (ConnectionResetError, BrokenPipeError):
                            log.warning("subscriber write error [%s] " % pipe.stream)
                            subscribers.remove(pipe)
                        except asyncio.TimeoutError:
                            log.warning("subscriber [%s] timed out" % pipe.stream)
                            pipe.close_interval_nowait()
                if child_output.end_of_interval:
                    for pipe in subscribers:
                        pipe.close_interval_nowait()

        except (EmptyPipe, asyncio.CancelledError):
            pass
        except PipeError as e:
            if 'closed pipe' in str(e):
                # during shutdown the pipe may be closed but
                # another read might be attempted by the output_handler
                pass
            else:
                log.warning("Worker %s, pipe %s: %s" % (
                    self.name, child_output.name, str(e)))

    async def _subscribe_to_inputs(self,
                                   subscribe: Callable[[Stream, pipes.Pipe, Loop], Callable],
                                   loop: Loop):
        # configure input pipes            [module]<==[worker]
        for (name, stream) in self.module.inputs.items():
            (r, w) = os.pipe()
            wf = pipes.writer_factory(w, loop)
            writer = await wf()
            os.set_inheritable(r, True)
            pipe = pipes.OutputPipe(name=name, stream=stream,
                                    writer=writer)
            try:
                unsubscribe = subscribe(stream, pipe, loop)
            except SubscriptionError as e:
                os.close(r)
                await pipe.close()
                raise e  # bubble exception up
            self.input_connections.append(DataConnection(name,
                                                         r, stream,
                                                         pipe,
                                                         unsubscribe))

    def _compose_cmd(self) -> str:
        cmd = shlex.split(self.module.exec_cmd)
        output_args = {}
        for c in self.output_connections:
            output_args[c.name] = {'fd': c.child_fd, 'id': c.stream.id, 'layout': c.stream.layout}
        input_args = {}
        for c in self.input_connections:
            input_args[c.name] = {'fd': c.child_fd, 'id': c.stream.id, 'layout': c.stream.layout}
        cmd += ["--pipes", json.dumps(json.dumps(
            {'outputs': output_args, 'inputs': input_args}))]
        # add a socket if the module has a web interface
        if self.module.is_app:
            cmd += ["--socket", self.interface_name]
            # remove the socket file if it exists
            if os.path.exists(self.interface_name):
                log.warning("forcibly removing app socket [%s]" % self.interface_name)
                os.unlink(self.interface_name)

        # API access
        cmd += ["--api-socket", API_SOCKET]
        for (arg, value) in self.module.arguments.items():
            cmd += ["--" + arg, value]
        return cmd

    def _close_child_fds(self):
        for c in self.input_connections + self.output_connections:
            os.close(c.child_fd)

    async def _close_connections(self):
        for c in self.output_connections + self.input_connections:
            await c.disconnect()
        self.output_connections = []
        self.input_connections = []
        # remove the socket file if it exists
        if self.interface_socket is not None and os.path.exists(self.interface_socket):
            os.unlink(self.interface_socket)

    def _verify_monotonic_timestamps(self, data, last_ts: int, name: str):
        if len(data) == 0:
            return True
        # if there are multiple rows, check that all timestamps are increasing
        if len(data) > 1 and np.min(np.diff(data['timestamp'])) <= 0:
            min_idx = np.argmin(np.diff(data['timestamp']))
            msg = ("Non-monotonic timestamp in new data to stream [%s] (%d<=%d)" %
                   (name, data['timestamp'][min_idx + 1], data['timestamp'][min_idx]))
            log.warning(msg)
            self.log(msg)
            return False
        # check to make sure the first timestamp is larger than the previous block
        if last_ts is not None:
            if last_ts >= data['timestamp'][0]:
                msg = ("Non-monotonic timestamp between writes to stream [%s] (%d<=%d)" %
                       (name, data['timestamp'][0], last_ts))
                log.warning(msg)
                self.log(msg)
                return False
        return True



import click
import subprocess
import configparser
import dsnparse
import os
import json
import asyncio
import shutil
from datetime import datetime
from joule import errors
import aiohttp


@click.command(name="backup")
@click.option("-c", "--config", help="main configuration file", default="/etc/joule/main.conf")
@click.option("-f", "--folder", help="backup folder name", default="joule_backup_NODE_DATE")
def admin_backup(config, folder):
    # expensive imports so only execute if the function is called
    from joule.services import load_config

    parser = configparser.ConfigParser()

    # load the Joule configuration file
    try:
        with open(config, 'r') as f:
            parser.read_file(f, config)
            config = load_config.run(custom_values=parser)
    except FileNotFoundError:
        raise click.ClickException("Cannot load joule configuration file at [%s]" % config)
    except PermissionError:
        raise click.ClickException("Cannot read joule configuration file at [%s] (run as root)" % config)
    except errors.ConfigurationError as e:
        raise click.ClickException("Invalid configuration: %s" % e)

    # demote priveleges
    if "SUDO_GID" in os.environ:
        os.setgid(int(os.environ["SUDO_GID"]))
    if "SUDO_UID" in os.environ:
        os.setuid(int(os.environ["SUDO_UID"]))

    if folder == "joule_backup_NODE_DATE":
        folder = "joule_backup_%s_%s" % (config.name, datetime.now().strftime("%Y%m%d_%H%M"))

    if os.path.exists(folder):
        raise click.ClickException("Requested folder [%s] already exists" % folder)

    os.mkdir(folder)

    # parse the dsn string
    parts = dsnparse.parse(config.database)

    # backup the database
    args = ["--format", "plain"]
    args += ["--checkpoint", "fast"]
    args += ["--pgdata", folder]
    args += ["--wal-method", "stream"]
    args += ["--progress"]
    args += ["--host", parts.host]
    args += ["--port", "%s" % parts.port]
    args += ["--username", parts.user]
    args += ["--label", "joule_backup"]
    cmd = ["pg_basebackup"] + args
    pg_proc_env = os.environ.copy()
    pg_proc_env["PGPASSWORD"] = parts.password
    click.echo("Copying up postgres database...")
    subprocess.call(cmd, env=pg_proc_env)

    # add the database name and user
    db_info = {
        "database": parts.database,
        "user": parts.user,
        "password": parts.secret,
        "nilmdb": config.nilmdb_url is not None
    }
    with open(os.path.join(folder, "info.json"), 'w') as f:
        f.write(json.dumps(db_info, indent=2))

    if config.nilmdb_url is None:
        click.echo("OK")
        return

    click.echo("Copying nilmdb database...")
    # retrieve the nilmdb data folder
    loop = asyncio.get_event_loop()
    nilmdb_folder = loop.run_until_complete(get_nilmdb_dir(config.nilmdb_url))
    nilmdb_backup = os.path.join(folder, "nilmdb")
    shutil.copytree(nilmdb_folder, nilmdb_backup, ignore=print_progress)
    click.echo("\nOK")


def print_progress(dir, contents):
    print('.', end="")
    return []


async def get_nilmdb_dir(url):
    async with aiohttp.ClientSession() as session:
        async with session.get(url + "/dbinfo") as resp:
            data = await resp.json()
            return data['path']



from joule.api.helpers import *
from joule.api.node import BaseNode
from joule.api.folder_type import Folder
from joule.api.stream import (Stream,
                              StreamInfo,
                              Element)
from joule.api.module import (Module,
                              ModuleStatistics)
from joule.api.annotation import Annotation
from joule.api.proxy import Proxy



import aiohttp
import ssl
from aiohttp.client_exceptions import ClientError
from typing import Optional


# parser for boolean args
def yesno(val: str):
    """
    Convert a "yes" or "no" argument into a boolean value. Returns ``true``
    if val is "yes" and ``false`` if val is "no". Raises ValueError otherwise.
    This is function can be used as the **type** parameter for to handle module arguments
    that are "yes|no" flags.
    """
    if val is None:
        raise ValueError("must be 'yes' or 'no'")
    # standardize the string
    val = val.lower().strip()
    if val == "yes":
        return True
    elif val == "no":
        return False
    else:
        raise ValueError("must be 'yes' or 'no'")


async def detect_url(host, port: Optional[int] = None):  # pragma: no cover
    if port is not None:
        host = host + ":" + str(port)

    ssl_context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
    ssl_context.check_hostname = False
    ssl_context.verify_mode = ssl.CERT_NONE

    async with aiohttp.ClientSession(conn_timeout=5) as session:
        # try to connect over https
        try:
            async with session.get("https://" + host, ssl_context=ssl_context) as resp:
                pass
            return "https://" + host
        except ClientError as e:
            # try again over http
            try:
                async with session.get("http://" + host) as resp:
                    pass
                return "http://" + host
            except ClientError:
                return None



import click
import asyncio

from joule import errors
from joule.cli.config import Config, pass_config
from joule.api import BaseNode


@click.command(name="rename")
@click.argument("stream")
@click.argument("name")
@pass_config
def cli_rename(config: Config, stream, name):
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            _run(config.node, stream, name))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()
    click.echo("OK")


async def _run(node: BaseNode, stream_path: str, name: str):
    stream = await node.stream_get(stream_path)
    stream.name = name
    await node.stream_update(stream)



import click
import asyncio

from joule import errors
from joule.api.stream import stream_delete
from joule.cli.config import pass_config


@click.command(name="delete")
@click.argument("stream")
@pass_config
def cli_delete(config, stream):
    if not click.confirm("Delete stream [%s]?" % stream):
        click.echo("Aborted!")
        return

    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            config.node.stream_delete(stream))
        click.echo("OK")

    except errors.ApiError as e:
        raise click.ClickException(str(e))
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()




import sqlalchemy
from urllib.parse import urlparse

from .session import BaseSession
from joule.utilities import connection_info


async def db_connect(session: BaseSession) -> sqlalchemy.engine.Engine:
    resp = await session.get("/db/connection.json")
    conn_info = connection_info.from_json(resp)
    # replace the host with the session host
    url = urlparse(session.url)
    conn_info.host = url.hostname
    return sqlalchemy.create_engine(conn_info.to_dsn())


async def db_connection_info(session: BaseSession) -> connection_info.ConnectionInfo:
    resp = await session.get("/db/connection.json")
    # replace the host with the session host
    conn_info = connection_info.from_json(resp)
    url = urlparse(session.url)
    conn_info.host = url.hostname
    return conn_info



import click
from .info import cli_info
from .list import cli_list
from .move import cli_move
from .destroy import cli_delete
from .annotation import cli_annotations
from .rename import cli_rename


@click.group(name="stream")
def streams():
    """Manage Joule data streams"""
    pass  # pragma: no cover


streams.add_command(cli_info)
streams.add_command(cli_list)
streams.add_command(cli_move)
streams.add_command(cli_delete)
streams.add_command(cli_annotations)
streams.add_command(cli_rename)



from joule.client.reader_module import ReaderModule
from joule.client.filter_module import FilterModule
from joule.client.composite_module import CompositeModule
from joule.client.base_module import BaseModule
from joule.models.pipes import LocalPipe, EmptyPipe, InputPipe, OutputPipe, Pipe
from joule.models import Stream, Element
from joule import utilities

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import click
from .initialize import admin_initialize
from .erase import admin_erase
from .authorize import admin_authorize
from .backup import admin_backup
from .ingest import admin_ingest


@click.group(name="admin")
def admin():
    pass  # pragma: no cover


admin.add_command(admin_initialize)
admin.add_command(admin_authorize)
admin.add_command(admin_erase)
admin.add_command(admin_backup)
admin.add_command(admin_ingest)



from joule.models.annotation import Annotation
from joule.models.element import Element
from joule.models.stream import Stream
from joule.models.folder import Folder
from joule.models.module import Module
from joule.models.worker import Worker
from joule.models.proxy import Proxy
from joule.models.master import Master
from joule.models.follower import Follower
from joule.models.data_store.data_store import DataStore, StreamInfo, DbInfo
from joule.models.data_store.errors import InsufficientDecimationError, DataError
from joule.models.data_store.nilmdb import NilmdbStore
from joule.models.data_store.timescale import TimescaleStore
from joule.models.pipes.pipe import Pipe
from joule.models.meta import Base



import click
import asyncio
import signal

from joule.cli.config import pass_config
from joule.models.pipes import EmptyPipe
from joule.api.data import (data_subscribe,
                            data_read)
from joule import errors
from joule.utilities import human_to_timestamp

stop_requested = False


@click.command(name="read")
@click.option('-s', "--start", help="timestamp or descriptive string")
@click.option('-e', "--end", help="timestamp or descriptive string")
@click.option('-l', "--live", is_flag=True, help="subscribe to an active stream")
@click.option('-r', "--max-rows", help="limit response data", type=int)
@click.option('-b', "--show-bounds", is_flag=True, help="include min/max for decimated data")
@click.option('-m', "--mark-intervals", help="include [# interval break] tags", is_flag=True)
@click.argument("stream")
@pass_config
def cmd(config, start, end, live, max_rows, show_bounds, mark_intervals, stream):
    signal.signal(signal.SIGTERM, handler)
    signal.signal(signal.SIGINT, handler)

    if live and (start is not None or end is not None):
        raise click.ClickException("specify either --live or --start/--end, not both")
    if live and max_rows is not None:
        raise click.ClickException("cannot specify --live and --max-rows")
    if start is not None:
        try:
            start = human_to_timestamp(start)
        except ValueError:
            raise click.ClickException("invalid start time: [%s]" % start)
    if end is not None:
        try:
            end = human_to_timestamp(end)
        except ValueError:
            raise click.ClickException("invalid end time: [%s]" % end)

    loop = asyncio.get_event_loop()

    async def _run():
        if live:
            pipe = await config.node.data_subscribe(stream)
        else:
            pipe = await config.node.data_read(stream, start, end, max_rows)
        try:
            while not stop_requested:
                try:
                    data = await asyncio.wait_for(pipe.read(flatten=True), 1)
                except asyncio.TimeoutError:
                    # check periodically for Ctrl-C (SIGTERM) even if server is slow
                    continue
                if pipe.decimated and not show_bounds:
                    # suppress the bound information
                    ncols = (data.shape[1] - 1) // 3 + 1
                    data = data[:, :ncols]
                pipe.consume(len(data))
                for row in data:
                    line = "%d %s" % (row[0], ' '.join('%f' % x for x in row[1:]))
                    click.echo(line)
                if pipe.end_of_interval and mark_intervals:
                    click.echo("# interval break")
        except EmptyPipe:
            pass
        await pipe.close()

    try:
        loop.run_until_complete(_run())
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


def handler(signum, frame):
    global stop_requested
    stop_requested = True



from sqlalchemy.orm import Session
from typing import List, Dict, TYPE_CHECKING
import logging

from joule.models import (Module, Stream)
from joule.errors import ConfigurationError
from joule.models.module import from_config as module_from_config

from joule.services import parse_pipe_config
from joule.services.helpers import (load_configs,
                                    Configurations)

if TYPE_CHECKING:
    from configparser import ConfigParser

logger = logging.getLogger('joule')

# types
Modules = List[Module]
StreamConnections = Dict[str, Stream]


def run(path: str, db: Session) -> Modules:
    configs = load_configs(path)
    return _parse_configs(configs, db)


def _parse_configs(configs: Configurations, db: Session) -> Modules:
    stage1_modules: List[(Module, ConfigParser, str)] = []
    # Pass 1: parse module parameters
    uuid = 0
    for file_path, config in configs.items():
        try:
            module = module_from_config(config)
            module.uuid = uuid
            stage1_modules.append((module, config, file_path))
            uuid += 1
        except ConfigurationError as e:
            logger.error("Invalid module [%s]: %s" % (file_path, e))
    # Pass 2: parse outputs
    stage2_modules: List[(Module, ConfigParser, str)] = []
    for module, config, file_path in stage1_modules:
        try:
            module.outputs = _connect_outputs(config, db)
            stage2_modules.append((module, config, file_path))
        except ConfigurationError as e:
            logger.error("Invalid module [%s]: %s" % (file_path, e))
    # Pass 3: parse inputs
    parsed_modules: Modules = []
    for module, config, file_path in stage2_modules:
        try:
            module.inputs = _connect_inputs(config, db)
            parsed_modules.append(module)
        except ConfigurationError as e:
            logger.error("Invalid module [%s]: %s" % (file_path, e))
    # Designate active streams
    for module in parsed_modules:
        for stream in module.outputs.values():
            stream.is_destination = True
        for stream in module.inputs.values():
            stream.is_source = True
    return parsed_modules


def _connect_outputs(config: 'ConfigParser', db: Session) -> StreamConnections:
    outputs: StreamConnections = {}
    if 'Outputs' not in config:
        return outputs
    for name, pipe_config in config['Outputs'].items():
        try:
            outputs[name] = parse_pipe_config.run(pipe_config, db)
        except ConfigurationError as e:
            raise ConfigurationError("Output [%s=%s]: %s" %
                                     (name, pipe_config, e))
    return outputs


def _connect_inputs(config: 'ConfigParser', db: Session) -> StreamConnections:
    inputs: StreamConnections = {}
    if 'Inputs' not in config:
        return inputs
    for name, pipe_config in config['Inputs'].items():
        try:
            inputs[name] = parse_pipe_config.run(pipe_config, db)
        except ConfigurationError as e:
            raise ConfigurationError("Input [%s=%s]: %s" %
                                     (name, pipe_config, e))
    return inputs



import click

from joule.cli.stream import streams
from joule.cli.data import data
from joule.cli.module import module
from joule.cli.folder import folders
from joule.cli.proxy import proxies
from joule.cli.admin import admin
from joule.cli.master import master
from joule.cli.follower import follower
from joule.cli.node import node
from joule.cli.config import Config, pass_config


@click.group()
@click.option('-n', '--node', default="", help="Joule Node name")
@click.version_option()
@pass_config
def main(config, node):
    # create a Node structure for the name
    # if node_name is given use it, otherwise go with the default
    config.set_node_name(node)


main.add_command(admin)
main.add_command(streams)
main.add_command(data)
main.add_command(module)
main.add_command(folders)
main.add_command(proxies)
main.add_command(master)
main.add_command(follower)
main.add_command(node)



import click
import os
import shutil
import subprocess
from typing import Tuple
from OpenSSL import crypto
from jinja2 import Environment, FileSystemLoader

import secrets


@click.command(name="initialize")
@click.option("--dsn", help="PostgreSQL DSN", required=True)
@click.option("--bind", help="IP address (0.0.0.0 for all)")
@click.option("--port", help="TCP port (default is 8080)")
@click.option("--nilmdb", help="NilmDB URL")
def admin_initialize(dsn, bind, port, nilmdb):  # pragma: no cover
    import pkg_resources

    # check arguments
    if bind is None and port is not None:
        raise click.ClickException("Must specify bind address with port")
    if bind is not None and port is None:
        port = "8080"
    click.echo("1. creating joule user ", nl=False)
    proc = subprocess.run("useradd -r -G dialout joule".split(" "), stderr=subprocess.PIPE)
    if proc.returncode == 0:
        click.echo("[" + click.style("OK", fg="green") + "]")
    elif proc.returncode == 1:
        _run_as_root()
    elif proc.returncode == 9:
        click.secho("[" + click.style("EXISTS", fg='yellow') + "]")
    else:
        click.echo("[" + click.style("ERROR", fg='red') + "]\n unknown error [%d] see [man useradd]" % proc.returncode)
        exit(1)
    click.echo("2. registering system service ", nl=False)
    service_file = pkg_resources.resource_filename(
        "joule", "resources/joule.service")
    try:
        shutil.copy(service_file, "/etc/systemd/system")
        subprocess.run("systemctl enable joule.service".split(" "), stderr=subprocess.PIPE)
        subprocess.run("systemctl start joule.service".split(" "), stderr=subprocess.PIPE)
    except PermissionError:
        _run_as_root()

    click.echo("[" + click.style("OK", fg="green") + "]")

    click.echo("3. copying configuration to /etc/joule ", nl=False)
    _make_joule_directory("/etc/joule")
    # check if main.conf exists
    if not os.path.isfile("/etc/joule/main.conf"):
        template_path = pkg_resources.resource_filename(
            "joule", "resources/templates")
        # From https://stackoverflow.com/questions/11857530
        env = Environment(loader=FileSystemLoader(template_path))
        env.trim_blocks = True
        env.lstrip_blocks = True
        env.keep_trailing_newline = True
        template = env.get_template('main.conf.jinja2')
        file_contents = template.render(name='node_%s' % secrets.token_hex(2),
                                        bind=bind,
                                        port=port,
                                        nilmdb=nilmdb,
                                        dsn=dsn)
        with open("/etc/joule/main.conf", 'w') as conf:
            conf.write(file_contents)

    # set ownership to joule user
    shutil.chown("/etc/joule/main.conf", user="root", group="joule")
    # only root can write, and only joule members can read
    os.chmod("/etc/joule/main.conf", 0o640)

    # setup stream config directory
    _make_joule_directory("/etc/joule/stream_configs")
    example_file = pkg_resources.resource_filename(
        "joule", "resources/templates/stream.example")
    shutil.copy(example_file, "/etc/joule/stream_configs")
    # set ownership to joule user
    shutil.chown("/etc/joule/stream_configs/stream.example",
                 user="joule", group="joule")

    # setup module config directory
    _make_joule_directory("/etc/joule/module_configs")
    example_file = pkg_resources.resource_filename(
        "joule", "resources/templates/module.example")
    shutil.copy(example_file, "/etc/joule/module_configs")
    # set ownership to joule user
    shutil.chown("/etc/joule/module_configs/module.example",
                 user="joule", group="joule")
    click.echo("[" + click.style("OK", fg="green") + "]")

    click.echo("4. creating SSL keys and certificates", nl=False)
    SECURITY_FOLDER = "/etc/joule/security"
    _make_joule_directory(SECURITY_FOLDER)
    shutil.chown(SECURITY_FOLDER, user="root", group="joule")
    os.chmod(SECURITY_FOLDER, 0o750)
    # check if a certificate file exists
    CERT_PATH = os.path.join(SECURITY_FOLDER, "server.crt")
    KEY_PATH = os.path.join(SECURITY_FOLDER, "server.key")
    if not os.path.isfile(KEY_PATH):
        cert, key = _self_signed_cert("joule_node")
        with open(CERT_PATH, "wb") as f:
            f.write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert))
        with open(KEY_PATH, "wb") as f:
            f.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, key))
        shutil.chown(CERT_PATH, user="root", group="joule")
        os.chmod(CERT_PATH, 0o640)
        shutil.chown(KEY_PATH, user="root", group="joule")
        os.chmod(KEY_PATH, 0o640)
        click.echo(" [" + click.style("OK", fg="green") + "]")
    else:
        click.echo(" [" + click.style("EXISTS", fg="yellow") + "]")


def _make_joule_directory(path):  # pragma: no cover
    try:
        if os.path.isfile(path):
            click.echo("\n " + click.style("ERROR", fg='red') +
                       " cannot create directory [%s], a file exists with the same name" % path)
            exit(1)
        # check if directory exists
        if not os.path.isdir(path):
            os.makedirs(path)
        # set ownership to joule user
        shutil.chown(path, user="joule", group="joule")
    except PermissionError:
        _run_as_root()


def _run_as_root():  # pragma: no cover
    click.echo("[" + click.style("ERROR", fg="red") + "]\n run as [sudo joule initialize]")
    exit(1)


def _self_signed_cert(cn: str) -> Tuple[crypto.X509, crypto.PKey]:
    # generate a key
    k = crypto.PKey()
    k.generate_key(crypto.TYPE_RSA, 2048)
    # generate a self signed certificate

    cert = crypto.X509()
    cert.get_subject().CN = cn
    cert.set_serial_number(secrets.randbits(128))
    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(5 * 365 * 24 * 60 * 60)  # 5 years
    cert.set_issuer(cert.get_subject())
    cert.set_pubkey(k)
    cert.sign(k, 'sha256')
    return cert, k



import aiohttp

from .base_session import BaseSession
from joule import errors


class UnixSession(BaseSession):

    def __init__(self, path: str, cafile: str):
        super().__init__()
        self.path = path
        self.url = "http://localhost"
        self.cafile = cafile

    async def get_session(self):
        if self._session is None:
            conn = aiohttp.UnixConnector(path=self.path)
            self._session = aiohttp.ClientSession(
                connector=conn,
                timeout=aiohttp.ClientTimeout(total=None))
        return self._session

    async def _request(self, method, path, data=None, json=None, params=None, chunked=None):
        session = await self.get_session()
        try:
            async with session.request(method,
                                       self.url + path,
                                       data=data,
                                       params=params,
                                       json=json) as resp:
                if resp.status != 200:
                    raise errors.ApiError("%s [%d]" % (await resp.text(),
                                                       resp.status))
                if resp.content_type != 'application/json':
                    body = await resp.text()
                    if body.lower() != "ok":
                        raise errors.ApiError("Invalid node response (not json)")
                    else:
                        return None
                try:
                    return await resp.json()
                except ValueError:
                    raise errors.ApiError("Invalid node response (not json)")

        except aiohttp.ClientError as e:
            raise errors.ApiError("Cannot contact node at [%s]" % self.url) from e



import click
import asyncio
from tabulate import tabulate

from joule import errors
from joule.cli.config import pass_config
from joule.api import BaseNode
from joule import utilities


@click.command(name="annotations")
@click.argument("stream")
@click.option('-s', "--start", help="timestamp or descriptive string", default=None)
@click.option('-e', "--end", help="timestamp or descriptive string", default=None)
@click.option("-c", "--csv", is_flag=True, help="display in raw csv format")
@click.option("--delete", is_flag=True, help="remove annotations for this stream")
@pass_config
def cli_annotations(config, stream, start, end, delete, csv):
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            _run(config.node, stream, start, end, delete, csv))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


async def _run(node: BaseNode, stream, start, end, delete, csv):
    if start is not None:
        try:
            start = utilities.human_to_timestamp(start)
        except ValueError:
            raise click.ClickException("invalid start time: [%s]" % start)
    if end is not None:
        try:
            end = utilities.human_to_timestamp(end)
        except ValueError:
            raise click.ClickException("invalid end time: [%s]" % end)
    if delete:
        await node.stream_annotation_delete(stream, start, end)
        click.echo("OK")
        return
    annotations = await node.annotation_get(stream, start, end)
    if csv:
        print_csv(annotations)
    else:
        print_table(annotations)


def print_csv(annotations):
    for a in annotations:
        if a.end is not None:
            print("%s,%s,%d,%d" % (a.title, a.content, a.start, a.end))
        else:
            print("%s,%s,%d" % (a.title, a.content, a.start))


def print_table(annotations):
    headers = ["Title", "Content", "Start", "End"]
    result = []
    for a in annotations:
        row = [a.title, a.content, utilities.timestamp_to_human(a.start)]
        if a.end is not None:
            row.append(utilities.timestamp_to_human(a.end))
        else:
            row.append("\u2014")
        result.append(row)
    click.echo(tabulate(result,
                        headers=headers,
                        tablefmt="fancy_grid"))



import click
import asyncio
from tabulate import tabulate

from joule import errors
from joule.api import BaseNode
from joule.cli.config import pass_config


@click.command(name="list")
@click.option('--statistics', '-s', is_flag=True, help="include memory and CPU statistics")
@pass_config
def cli_list(config, statistics):
    loop = asyncio.get_event_loop()

    try:
        loop.run_until_complete(
            _run(config.node, statistics))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


async def _run(node: BaseNode, statistics):
    modules = await node.module_list(statistics)
    # display module information
    headers = ['Name', 'Inputs', 'Outputs']
    if statistics:
        headers += ['CPU %', "Mem %"]
    result = []
    for module in modules:
        inputs = '\n'.join(module.inputs.values())
        outputs = '\n'.join(module.outputs.values())
        data = [module.name, inputs, outputs]
        if statistics:
            if module.statistics.cpu_percent is None:
                cpu_stat = "\u2014"
            else:
                cpu_stat = "%0.1f" % module.statistics.cpu_percent
            if module.statistics.memory_percent is None:
                mem_stat = "\u2014"
            else:
                mem_stat = '%0.1f' % module.statistics.memory_percent
            data += [cpu_stat, mem_stat]
        result.append(data)
    click.echo(tabulate(result,
                        headers=headers,
                        tablefmt="fancy_grid"))



import click

from joule import api


class Config:
    def __init__(self):
        self._node: api.BaseNode = None
        self.name = ""

    def set_node_name(self, name):
        self.name = name

    @property
    def node(self):
        # lazy node construction, raise error if it cannot be created
        if self._node is None:
            self._node = api.get_node(self.name)
            click.echo("--connecting to [%s]--" % self._node.name, err=True)
            self.name = self._node.name
        return self._node

    async def close_node(self):
        if self._node is None:
            return
        else:
            await self._node.close()


pass_config = click.make_pass_decorator(Config, ensure=True)



from typing import List, TYPE_CHECKING
from joule import errors

if TYPE_CHECKING:
    from .stream import Stream


class Folder:
    """
    API Folder model. See :ref:`sec-node-folder-actions` for details on using the API to
    manipulate folders. Folders are locked if they contain locked streams (streams that are active
    or statically configured). Folder objects should not be created directly. They should be received
    via API calls.

    Parameters:
        name (str): folder name, must be unique in the parent
        description (str): optional field
        locked (bool): folder may not be moved, deleted, or changed
        streams (List[Stream]): streams in the folder
        children (List[Folder]): subfolders in the folder

    """
    # folder params
    def __init__(self):
        self._id = None
        self.name: str = ""
        self.description: str = ""
        self.locked: bool = False
        self.streams: List['Stream'] = []
        self.children: List['Folder'] = []

    def __repr__(self):
        return "<joule.api.Folder id=%r name=%r description=%r locked=%r>" % (
                self._id, self.name, self.description, self.locked)

    @property
    def id(self) -> int:
        if self._id is None:
            raise errors.ApiError("this is a local model with no ID. See API docs")
        return self._id

    @id.setter
    def id(self, value: int):
        self._id = value

    def to_json(self):
        return {
            "id": self._id,
            "name": self.name,
            "description": self.description,
            "streams": [s.to_json() for s in self.streams],
            "children": [c.to_json() for c in self.children],
            "locked": self.locked
        }



#!/usr/bin/env python

from setuptools import setup, find_packages
import versioneer

PROJECT = 'Joule'

# Change docs/sphinx/conf.py too!

try:
    long_description = open('README.rst', 'rt').read()
except IOError:
    long_description = ''

setup(
    name=PROJECT,
    version = versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Decentralized data processing for IoT platforms',
    long_description=long_description,

    author='John Donnal',
    author_email='donnal@usna.edu',

    url='http://wattsworth.net/joule',
    download_url='https://github.com/wattsworth/joule',
    license='open source (see LICENSE)',
    classifiers=['Programming Language :: Python',
                 'Environment :: Console',
                 ],
    platforms=['Any'],
    scripts=[],
    provides=[],
    install_requires=['click',
                      'treelib',
                      'numpy',
                      'scipy',
                      'psutil',
                      'aiohttp',
                      'markdown',
                      'BeautifulSoup4',
                      'dateparser',
                      'tabulate',
                      'sqlalchemy',
                      'aiohttp-jinja2',
                      'jinja2',
                      'asyncpg',
                      'psycopg2-binary',
                      'uvloop',
                      'aiodns',
                      'cchardet',
                      'pyopenssl',
                      'dsnparse'],
    tests_require=['asynctest',
                   'nose2',
                   'testing.postgresql',
                   'requests'],
    test_suite='tests',
    namespace_packages=[],
    packages=find_packages(exclude=["tests","tests.*"]),
    include_package_data=True,

    entry_points={
        'console_scripts': [
            'joule = joule.cli:main',
            'jouled = joule.daemon:main',
            'joule-random-reader = joule.client.builtins.random_reader:main',
            'joule-file-reader = joule.client.builtins.file_reader:main',
            'joule-mean-filter = joule.client.builtins.mean_filter:main',
            'joule-median-filter = joule.client.builtins.median_filter:main',
            'joule-visualizer-filter = joule.client.builtins.visualizer:main',
            'joule-merge-filter = joule.client.builtins.merge_filter:main'
        ]
    },
    #options={
    #    'build_scripts': {
    #        'executable': '/usr/local/bin/python3.5'
    #    }
    #},
    zip_safe=False,
)



from typing import Union, List, Optional, TYPE_CHECKING

from joule import errors
from .session import BaseSession

if TYPE_CHECKING:
    from .stream import Stream


class Annotation:
    """
    API Annotation model. See :ref:`sec-node-annotation-actions` for details on using the API to
    manipulate annotations. Annotations are associated with streams and may either coverage a range
    of data or a single event. If **end** is ``None`` the annotation marks an event, otherwise
    it marks a range.

    Parameters:
       title (string): annotation title
       content (string): additional description (optional)
       start (int): Unix microsecond timestamp
       end (int): specify for range annotation, omit for event annotation
    """

    def __init__(self, title: str,
                 start: int,
                 end: Optional[int] = None,
                 content: str = ""):
        self.id = None
        self.title = title
        self.content = content
        self.stream_id = None
        self.start = start
        self.end = end

    def __eq__(self, other: 'Annotation'):
        if self.title != other.title:
            return False
        if self.content != other.content:
            return False
        if self.start != other.start:
            return False
        if self.end != other.end:
            return False
        return True

    def to_json(self):
        return {
            'title': self.title,
            'start': self.start,
            'end': self.end,
            'content': self.content,
            'id': self.id
        }


def from_json(json) -> Annotation:
    annotation = Annotation(title=json['title'],
                            start=json['start'],
                            end=json['end'],
                            content=json['content'])
    annotation.id = json["id"]
    annotation.stream_id = json["stream_id"]
    return annotation


async def annotation_create(session: BaseSession,
                            annotation: Annotation,
                            stream: Union[int, str, 'Stream'], ) -> Annotation:
    from .stream import Stream

    data = {"title": annotation.title,
            "content": annotation.content,
            "start": annotation.start,
            "end": annotation.end}
    if type(stream) is Stream:
        data["stream_id"] = stream.id
    elif type(stream) is int:
        data["stream_id"] = stream
    elif type(stream) is str:
        data["stream_path"] = stream
    else:
        raise errors.ApiError("Invalid stream datatype. Must be Stream, Path, or ID")

    json = await session.post("/annotation.json", json=data)
    return from_json(json)


async def annotation_delete(session: BaseSession,
                            annotation: Union[int, Annotation]):
    data = {}
    if type(annotation) is Annotation:
        data["id"] = annotation.id
    elif type(annotation) is int:
        data["id"] = annotation

    await session.delete("/annotation.json", params=data)


async def annotation_update(session: BaseSession,
                            annotation: Annotation):
    data = {"id": annotation.id,
            "title": annotation.title,
            "content": annotation.content,
            }
    json = await session.put("/annotation.json", json=data)
    return from_json(json)


async def annotation_get(session: BaseSession,
                         stream: Union['Stream', str, int],
                         start: Optional[int],
                         end: Optional[int]) -> List[Annotation]:
    from .stream import Stream

    data = {}
    if start is not None:
        data["start"] = start
    if end is not None:
        data["end"] = end

    if type(stream) is Stream:
        data["stream_id"] = stream.id
    elif type(stream) is int:
        data["stream_id"] = stream
    elif type(stream) is str:
        data["stream_path"] = stream
    else:
        raise errors.ApiError("Invalid stream datatype. Must be Stream, Path, or ID")

    json = await session.get("/annotations.json", data)

    annotations = []
    for item in json:
        annotations.append(from_json(item))
    return annotations



import click
import asyncio

from joule.cli.config import pass_config
from joule.api.data import data_delete
from joule.utilities import human_to_timestamp
from joule import errors


@click.command(name="delete")
@click.option("-s", "--start", "start", help="timestamp or descriptive string")
@click.option("-e", "--end", "end", help="timestamp or descriptive string")
@click.option("--all", is_flag=True, help="remove all data")
@click.argument("stream")
@pass_config
def data_remove(config, start, end, all, stream):
    loop = asyncio.get_event_loop()
    if all:
        if start is not None or end is not None:
            raise click.ClickException("specify either --all or --start/--end")
    if all is None and start is None and end is None:
        raise click.ClickException("specify either --all or --start/--end")
    if start is not None:
        try:
            start = human_to_timestamp(start)
        except ValueError:
            raise click.ClickException("invalid start time: [%s]" % start)
    if end is not None:
        try:
            end = human_to_timestamp(end)
        except ValueError:
            raise click.ClickException("invalid end time: [%s]" % end)

    try:
        loop.run_until_complete(config.node.data_delete(
            stream, start, end))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    except:
        import traceback
        traceback.print_exc()
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()
    click.echo("OK")



from typing import List, Union, Dict

from joule import errors
from . import stream
from .session import BaseSession


class ModuleStatistics:
    """
    API ModuleStatistics model. Received by settings statistics to ``True`` when calling :meth:`Node.module_get` or
    :meth:`Node.module_info` and should not be created directly.

    Parameters:
        pid (int): process ID
        create_time (int): process creation time in UNIX seconds
        cpu_percent (float): approximate CPU usage
        memory_percent (float): approximate memory usage
    """

    def __init__(self, pid: int, create_time: int,
                 cpu_percent: float, memory_percent: float):
        self.pid = pid
        self.create_time = create_time
        self.cpu_percent = cpu_percent
        self.memory_percent = memory_percent

    def __repr__(self):
        return "<joule.api.ModuleStatistics pid=%r create_time=%r cpu_percent=%r memory_percent=%r>" % (
            self.pid, self.create_time, self.cpu_percent, self.memory_percent)


class Module:
    """
    API Module model. See :ref:`sec-node-module-actions` for details on using the API to
    query modules. See :ref:`modules` for details on writing new modules. See :ref:`sec-modules`
    for details on adding modules to Joule. Use :meth:`Node.stream_get` to retrieve the associated stream from the path string
    in the ``inputs`` and ``outputs`` dictionaries.

    Parameters:
       id (int): unique numeric ID assigned by Joule server
       name (str): module name, must be unique
       description (str): optional field
       is_app (bool): whether the module provides a web interface
       inputs (Dict): Mapping of ``[pipe_name] = stream_path`` for module inputs
       outputs (Dict): Mapping of ``[pipe_name] = stream_path`` for output connections
       statistics (ModuleStatistics): execution statistics, may be ``None`` depending on API call parameters
    """

    def __init__(self, id: int, name: str, description: str,
                 is_app: bool,
                 inputs: Dict[str, stream.Stream],
                 outputs: Dict[str, stream.Stream],
                 statistics: ModuleStatistics = None):
        self.id = id
        self.name = name
        self.description = description
        self.is_app = is_app
        self.inputs = inputs
        self.outputs = outputs
        self.statistics = statistics

    def __repr__(self):
        return "<joule.api.Module id=%r name=%r description=%r is_app=%r>" % (
            self.id, self.name, self.description, self.is_app)


def from_json(json) -> Module:
    inputs = json['inputs']
    outputs = json['outputs']

    if 'statistics' in json:
        s = json['statistics']
        statistics = ModuleStatistics(s['pid'], s['create_time'],
                                      s['cpu_percent'],
                                      s['memory_percent'])
    else:
        statistics = None

    return Module(json['id'], json['name'], json['description'],
                  json['is_app'],
                  inputs, outputs, statistics)


async def module_get(session: BaseSession,
                     module: Union[Module, str, int],
                     statistics: bool = True) -> Module:
    _statistics = 0
    if statistics:
        _statistics = 1

    params = {"statistics": _statistics}
    if type(module) is Module:
        params["id"] = module.id
    elif type(module) is str:
        params["name"] = module
    elif type(module) is int:
        params["id"] = module
    else:
        raise errors.ApiError("Invalid module datatype. Must be Module, Name, or ID")

    resp = await session.get("/module.json", params)
    return from_json(resp)


async def module_list(session: BaseSession,
                      statistics: bool = False) -> List[Module]:
    _statistics = 1
    if not statistics:
        _statistics = 0
    resp = await session.get("/modules.json", {"statistics": _statistics})
    return [from_json(item) for item in resp]


async def module_logs(session: BaseSession,
                      module: Union[Module, str, int]) -> List[str]:
    params = {}
    if type(module) is Module:
        params["id"] = module.id
    elif type(module) is str:
        params["name"] = module
    elif type(module) is int:
        params["id"] = module
    else:
        raise errors.ApiError("Invalid module datatype. Must be Module, Name, or ID")

    return await session.get("/module/logs.json", params)



import click
import os
import configparser

from joule import api, errors


@click.command(name="authorize")
@click.option("-c", "--config", help="main configuration file", default="/etc/joule/main.conf")
@click.option("-u", "--url", help="joule API URL")
def admin_authorize(config, url):
    # expensive imports so only execute if the function is called
    from joule.services import load_config
    from joule.models import (Base, master)
    from sqlalchemy import create_engine
    from sqlalchemy.orm import Session

    parser = configparser.ConfigParser()

    # load the Joule configuration file
    try:
        with open(config, 'r') as f:
            parser.read_file(f, config)
            config = load_config.run(custom_values=parser)
    except FileNotFoundError:
        raise click.ClickException("Cannot load joule configuration file at [%s]" % config)
    except PermissionError:
        raise click.ClickException("Cannot read joule configuration file at [%s] (run as root)" % config)
    except errors.ConfigurationError as e:
        raise click.ClickException("Invalid configuration: %s" % e)

    # create a connection to the database
    engine = create_engine(config.database)
    with engine.connect() as conn:
        conn.execute('CREATE SCHEMA IF NOT EXISTS data')
        conn.execute('CREATE SCHEMA IF NOT EXISTS metadata')

    Base.metadata.create_all(engine)
    db = Session(bind=engine)

    if 'SUDO_USER' in os.environ:
        username = os.environ["SUDO_USER"]
    else:
        username = os.environ["LOGNAME"]

    try:
        nodes = api.get_nodes()
    except ValueError as e:
        raise click.ClickException(str(e))

    # check if this name is associated with a master entry
    my_master = db.query(master.Master). \
        filter(master.Master.type == master.Master.TYPE.USER). \
        filter(master.Master.name == username).first()
    if my_master is None:
        # create a new master entry
        my_master = master.Master()
        my_master.key = master.make_key()
        my_master.type = master.Master.TYPE.USER
        my_master.name = username
        db.add(my_master)

    # if a url is specified use it
    if url is not None:
        joule_url = url
    # if the Joule server is not hosting a TCP server use the
    # default proxy address
    elif config.ip_address is None:
        joule_url = "https://localhost/joule"
    # otherwise use the the server information in the config file
    else:
        if config.security is not None and config.security.cafile != "":
            addr = config.name
        elif config.ip_address != "0.0.0.0":
            addr = config.ip_address
        else:
            addr = "127.0.0.1"

        if config.security is None:
            scheme = "http"
        else:
            scheme = "https"
        joule_url = "%s://%s:%d" % (scheme, addr, config.port)
    my_node = api.create_tcp_node(joule_url, my_master.key, config.name)
    api.save_node(my_node)

    db.commit()
    db.close()

    click.echo("Access to node [%s] granted to user [%s]" % (
        config.name, username))
